/* ++++++++++
	driver.c
	A skeletal device driver
+++++ */

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <ISA.h>
#include "driverdef.h"

static isa_module_info *isa;
bool occured = false;
pardev the_port; // the global port

/* ----------
	init_hardware - called once the first time the driver is loaded
----- */
status_t
init_hardware (void)
{
	return B_OK;
}
 

static int32 handler(void *data)
{
	pardev *d = (pardev *)data;
	int32 semcount;
	
	d->intoccured = false;
	
	if(occured == false) occured = true;
	else occured = false;
	
	if(occured == true) {
		
		// protect this with disable() !!
		isa->write_io_8(0x20, 0x20);
		
		
		dprintf("it's occured\n");
		// release all the threads blocked at this time on the sem (not only one !)
		get_sem_count(d->releaser1, &semcount);
		if(semcount < 0)
			release_sem_etc(d->releaser1, -semcount, B_DO_NOT_RESCHEDULE);//par martin
		
		return B_HANDLED_INTERRUPT;
	} else {
		dprintf("it hasn't occured\n");
	}
	
	d->intoccured = occured;
	
	return d->intoccured == 1 ? true : false;
}
/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */
status_t
init_driver (void)
{
	int int_mask;
	pardev *d = &the_port; // we put those things in the global...
	
	status_t s = get_module(B_ISA_MODULE_NAME, (module_info **)&isa);
	
	if(s != B_OK)
		goto ERROR_1;
	s = d->mainsem = create_sem(1, "aliensem");
	if(s < B_OK)
		goto ERROR_2;
	s = d->releaser1 = create_sem(0, "IRQ 7 releaser");// test par martin
	if(s < B_OK)
		goto ERROR_3;

	d->irq = 7;// + 0x08;
		
	s = acquire_sem(d->mainsem);
	if(s < B_OK)
		goto ERROR_4;

		
	/* mask interrupts */
//	int_mask=isa->read_io_8(0x21);
//    isa->write_io_8(0x21, int_mask | 0x80); 
    isa->write_io_8(CONTROL, isa->read_io_8(CONTROL) & ~0x10);

	/* it's better to install the handler after having masked the ints,
	 * else we can get spurious ints...
	 */
	if(s = install_io_interrupt_handler( d->irq, handler, d, 0) < B_OK)
		goto ERROR_4;

	release_sem(d->mainsem);

	return B_OK;
	
// in case of error here, we must remease all 
// the resource we got until the error occured

ERROR_4:
	delete_sem(d->releaser1);
ERROR_3:
	delete_sem(d->mainsem);
ERROR_2:
	put_module(B_ISA_MODULE_NAME);
ERROR_1:
	return(s);
}


/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */
void
uninit_driver (void)
{
	int int_mask;
	pardev *d = &the_port;

	acquire_sem(d->mainsem);

	// mask ints
//	int_mask=isa->read_io_8(0x21);
//	isa->write_io_8(0x21, int_mask | 0x80); 
    isa->write_io_8(CONTROL, isa->read_io_8(CONTROL) & ~0x10);

	// uninstall handler
	remove_io_interrupt_handler( d->irq, handler, d); 

	release_sem(d->mainsem);
	
	// clean up
	delete_sem(d->mainsem);
	delete_sem(d->releaser1);
	put_module(B_ISA_MODULE_NAME);
}

	
/* ----------
	my_device_open - handle open() calls
----- */

static status_t
my_device_open (const char *name, uint32 flags, void** cookie)
{
	pardev *d;
	int int_mask;
	
	d = (pardev *)malloc(sizeof(pardev));
	
	memcpy(d, &the_port, sizeof(pardev));	// copy useful infos from the global
											// like the sem ids...
	
	*cookie = (void *)d;	// assign this to the cookie
	
	acquire_sem(d->mainsem);
	
	// unmask ints
//	int_mask=isa->read_io_8(0x21);
//	isa->write_io_8(0x21, int_mask & ~0x80); 
    isa->write_io_8(CONTROL, isa->read_io_8(CONTROL) | 0x10);
	
	release_sem(d->mainsem);
	return B_OK;
}


/* ----------
	my_device_read - handle read() calls
----- */

static status_t
my_device_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	*num_bytes = 0;				/* tell caller nothing was read */
	return B_OK;
}


/* ----------
	my_device_write - handle write() calls
----- */

static status_t
my_device_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	size_t left = *num_bytes;
	while(left--)
		isa->write_io_8(DATA, *(char *)buffer++);
	*num_bytes;				/* tell caller nothing was written */
	return B_OK;
}


/* ----------
	my_device_control - handle ioctl calls
----- */

static status_t
my_device_control (void* cookie, uint32 op, void* arg, size_t len)
{
	pardev *iot = (pardev *)arg;
	thread_info thi;
  switch (op)
    {
        case INPORTB:
          	iot->Data = isa->read_io_8(iot->Port); //inportb implementation
          return B_OK;

        case OUTPORTB:
          	isa->write_io_8(iot->Port, iot->Data); //outportb implementation
          return B_OK;
          
        case INPORTW:
         	iot->Data16 = isa->read_io_16(iot->Port); //inportw implementation
         	return B_OK;
         
         case OUTPORTW:
         	isa->write_io_16(iot->Port, iot->Data16); //outportw implementation
         	return B_OK;
         case OCCUR:
         	iot->intoccured = occured;
         	return B_OK;
         case GET_INT_SEM:
         	iot->releaser1 = the_port.releaser1;
         	// Bad way...
         	get_thread_info(find_thread(NULL), &thi);
         	set_sem_owner(the_port.releaser1, thi.team);

        default:
          return B_BAD_VALUE;
      }
	return B_BAD_VALUE;
}



/* ----------
	my_device_close - handle close() calls
----- */

static status_t
my_device_close (void* cookie)
{
	
	return B_OK;
}


/* -----
	my_device_free - called after the last device is closed, and after
	all i/o is complete.
----- */
static status_t
my_device_free (void* cookie)
{
	pardev *d = (pardev *)cookie;
	int int_mask;
	
	acquire_sem(d->mainsem);
	
	int_mask=isa->read_io_8(0x21);
    isa->write_io_8(0x21, int_mask | 0x80);
    isa->write_io_8(CONTROL, isa->read_io_8(CONTROL) & ~0x10);
	
	release_sem(d->mainsem);

	return B_OK;
}


/* -----
	null-terminated array of device names supported by this driver
----- */

static const char *my_device_name[] = {
	"misc/alienparallel",
	NULL
};

/* -----
	function pointers for the device hooks entry points
----- */

device_hooks my_device_hooks = {
	my_device_open, 			/* -> open entry point */
	my_device_close, 			/* -> close entry point */
	my_device_free,			/* -> free cookie */
	my_device_control, 		/* -> control entry point */
	my_device_read,			/* -> read entry point */
	my_device_write			/* -> write entry point */
};

/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */

const char**
publish_devices()
{
	return my_device_name;
}

/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */

device_hooks*
find_device(const char* name)
{
	return &my_device_hooks;
}

